import featureflow as ff
from random_samples import ShuffledSamples
from random_samples import InfiniteSampler
from preprocess import PreprocessingPipeline


def learning_pipeline():
    class LearningPipeline(ff.BaseModel):
        samples = ff.PickleFeature(ff.IteratorNode)

        shuffled = ff.PickleFeature(
            ShuffledSamples,
            nsamples=ff.Var('nsamples'),
            dtype=ff.Var('dtype'),
            needs=samples)

    return LearningPipeline


def infinite_streaming_learning_pipeline(cls):
    roots = filter(lambda feature: feature.is_root, cls.features.itervalues())

    if len(roots) != 1:
        raise ValueError('cls must have a single root feature')

    root = roots[0]

    class InfiniteLearningPipeline(cls):
        dataset = ff.Feature(
            InfiniteSampler,
            nsamples=ff.Var('nsamples'),
            dtype=ff.Var('dtype'))

        pipeline = ff.ClobberPickleFeature(
            PreprocessingPipeline,
            needs=cls.features,
            store=True)

        @classmethod
        def load_network(cls):
            if not cls.exists():
                raise RuntimeError('No network has been trained or saved')

            instance = cls()
            for p in instance.pipeline:
                try:
                    return p.network
                except AttributeError:
                    pass

            raise RuntimeError('There is no network in the pipeline')

    root.needs = InfiniteLearningPipeline.dataset
    InfiniteLearningPipeline.__name__ = cls.__name__
    InfiniteLearningPipeline.__module__ = cls.__module__

    return InfiniteLearningPipeline



import unittest2
from zounds.util import simple_in_memory_settings
import featureflow as ff
from preprocess import InstanceScaling, PreprocessingPipeline
import numpy as np


class InstanceScalingTests(unittest2.TestCase):

    def get_model(self, max_value=1):
        @simple_in_memory_settings
        class Model(ff.BaseModel):
            scaled = ff.PickleFeature(
                InstanceScaling,
                max_value=max_value,
                store=False)

            pipeline = ff.PickleFeature(
                PreprocessingPipeline,
                needs=(scaled,),
                store=True)

        training = np.random.random_sample((10, 3))
        _id = Model.process(scaled=training)
        return Model(_id)

    def test_can_scale_to_arbitrary_value(self):
        model = self.get_model(max_value=50)
        inp = np.random.random_sample((100, 30)) * 10
        transformed = model.pipeline.transform(inp).data
        self.assertEqual(50.0, transformed.max())

    def test_can_invert_when_scaling_to_arbitrary_value(self):
        model = self.get_model(max_value=50)
        inp = np.random.random_sample((100, 30)) * 10
        transformed = model.pipeline.transform(inp)
        inverted = transformed.inverse_transform()
        np.testing.assert_allclose(inverted, inp)

    def test_forward_transform_scales_data_2d(self):
        model = self.get_model()
        inp = np.random.random_sample((100, 30)) * 10
        transformed = model.pipeline.transform(inp).data
        self.assertEqual(1.0, transformed.max())

    def test_each_example_has_same_max(self):
        model = self.get_model()
        inp = np.random.normal(0, 1, (100, 30)) * 50
        transformed = model.pipeline.transform(inp).data
        np.testing.assert_allclose(np.abs(transformed).max(axis=-1), 1)

    def test_forward_transform_scales_data_3d(self):
        model = self.get_model()
        inp = np.random.random_sample((100, 30, 3)) - 0.5
        transformed = model.pipeline.transform(inp).data
        self.assertEqual(1.0, np.max(np.abs(transformed)))

    def test_backward_transform_reconstructs_data_2d(self):
        model = self.get_model()
        inp = np.random.random_sample((100, 30)) * 10
        transformed = model.pipeline.transform(inp)
        inverted = transformed.inverse_transform()
        np.testing.assert_allclose(inverted, inp)

    def test_backward_transform_reconstructs_data_3d(self):
        model = self.get_model()
        inp = np.random.random_sample((100, 30, 3)) - 0.5
        transformed = model.pipeline.transform(inp)
        inverted = transformed.inverse_transform()
        np.testing.assert_allclose(inverted, inp)

    def test_correctly_handles_max_of_zero(self):
        model = self.get_model()
        inp = np.random.random_sample((100, 30, 3)) - 0.5
        inp[0, ...] = 0
        transformed = model.pipeline.transform(inp)
        inverted = transformed.inverse_transform()
        np.testing.assert_allclose(inverted, inp)

    def test_instance_scaling_maintains_dtype(self):
        model = self.get_model()
        inp = np.random.random_sample((100, 30)) * 10
        inp = inp.astype(np.float32)
        transformed = model.pipeline.transform(inp).data
        self.assertEqual(np.float32, transformed.dtype)



import os


class Directory(object):
    def __init__(self, path):
        self.path = path

    def __iter__(self):
        for f in os.listdir(self.path):
            p = os.path.join(self.path, f)
            if not os.path.isdir(p):
                yield os.path.join(self.path, f)



from featureflow import Feature
from zounds.persistence.arraywithunits import \
    ArrayWithUnitsDecoder, ArrayWithUnitsEncoder
from zounds.timeseries import audio_sample_rate, AudioSamples


class AudioSamplesDecoder(ArrayWithUnitsDecoder):
    def __init__(self):
        super(ArrayWithUnitsDecoder, self).__init__()

    def __call__(self, flo):
        raw = super(AudioSamplesDecoder, self).__call__(flo)
        samplerate = audio_sample_rate(raw.dimensions[0].samples_per_second)
        return AudioSamples(raw, samplerate)


class AudioSamplesFeature(Feature):

    def __init__(
            self,
            extractor,
            needs=None,
            store=False,
            key=None,
            encoder=ArrayWithUnitsEncoder,
            decoder=AudioSamplesDecoder(),
            **extractor_args):
        super(AudioSamplesFeature, self).__init__(
                extractor,
                needs=needs,
                store=store,
                encoder=encoder,
                decoder=decoder,
                key=key,
                **extractor_args)



import os
import re
import tornado.ioloop
import tornado.web
import httplib
import urllib
from contentrange import RangeRequest, UnsatisfiableRangeRequestException
from serializer import DefaultSerializer, AudioSamplesSerializer, \
    NumpySerializer, OggVorbisSerializer, ConstantRateTimeSeriesSerializer, \
    OnsetsSerializer, SearchResultsSerializer
import threading
from uuid import uuid4


class NoMatchingSerializerException(Exception):
    pass


class RequestContext(object):
    def __init__(
            self,
            document=None,
            feature=None,
            slce=None,
            value=None):
        self.value = value
        self.slce = slce
        self.feature = feature
        self.document = document

    def __repr__(self):
        return '''RequestContext(
    document={document},
    feature={feature},
    slce={slce},
    value={value})'''.format(**self.__dict__)

    def __str__(self):
        return self.__repr__()


class BaseZoundsApp(object):
    def __init__(
            self,
            base_path=r'/zounds/',
            model=None,
            visualization_feature=None,
            audio_feature=None,
            html=None,
            secret=None):

        super(BaseZoundsApp, self).__init__()
        self.secret = secret
        self.locals = locals
        self.globals = globals
        self.model = model
        self.visualization_feature = visualization_feature
        self.audio_feature = audio_feature
        self.base_path = base_path
        self.serializers = [
            AudioSamplesSerializer(),
            OggVorbisSerializer(),
            ConstantRateTimeSeriesSerializer(),
            DefaultSerializer('application/json'),
            DefaultSerializer('audio/ogg'),
            NumpySerializer(),
            OnsetsSerializer(
                self.visualization_feature,
                self.audio_feature,
                self.feature_path),
            SearchResultsSerializer(
                self.visualization_feature,
                self.audio_feature,
                self.feature_path)
        ]
        self._html_content = self._get_html(html)
        self.server = None
        self.thread = None

    SCRIPT_TAG = re.compile('<script\s+src="/(?P<filename>[^"]+)"></script>')

    def _get_html(self, html_filename):
        path, fn = os.path.split(__file__)
        with open(os.path.join(path, html_filename)) as f:
            html = f.read()

        to_replace = [(m.groupdict()['filename'], html[m.start(): m.end()])
                      for m in self.SCRIPT_TAG.finditer(html)]

        for filename, tag in to_replace:
            with open(os.path.join(path, filename)) as scriptfile:
                html = html.replace(
                    tag, '<script>{}</script>'.format(scriptfile.read()))

        return html

    def feature_path(self, _id, feature):
        _id = urllib.quote(_id, safe='')
        return '{base_path}{_id}/{feature}'.format(
            base_path=self.base_path, _id=_id, feature=feature)

    def find_serializer(self, context):
        try:
            return filter(
                lambda x: x.matches(context), self.serializers)[0]
        except IndexError:
            raise NoMatchingSerializerException()

    def serialize(self, context):
        serializer = self.find_serializer(context)
        return serializer.serialize(context)

    def feature_handler(self):

        document = self.model
        app = self

        class FeatureHandler(tornado.web.RequestHandler):

            def get(self, _id, feature):
                doc = document(_id)
                feature = document.features[feature]
                try:
                    slce = RangeRequest(self.request.headers['Range']).range()
                except KeyError:
                    slce = slice(None)
                context = RequestContext(
                    document=doc, feature=feature, slce=slce)
                try:
                    result = app.serialize(context)
                except UnsatisfiableRangeRequestException:
                    self.set_status(httplib.REQUESTED_RANGE_NOT_SATISFIABLE)
                    self.finish()
                self.set_header('Content-Type', result.content_type)
                self.set_header('Accept-Ranges', 'bytes')
                self.write(result.data)
                self.set_header('ETag', self.compute_etag())
                self.set_status(
                    httplib.PARTIAL_CONTENT if result.is_partial
                    else httplib.OK)
                if result.content_range:
                    self.set_header('Content-Range', str(result.content_range))
                self.finish()

        return FeatureHandler

    def ui_handler(self):
        app = self

        class UIHandler(tornado.web.RequestHandler):
            def get(self):
                self.set_header('Content-Type', 'text/html')
                self.write(app._html_content)
                self.set_status(httplib.OK)
                self.finish()

        return UIHandler

    def login_handler(self):
        app = self

        class LoginHandler(tornado.web.RequestHandler):
            def get(self):
                self.write(app._get_html('login.html'))
                self.finish()

            def post(self):
                secret = self.get_body_argument('secret')
                if secret == app.secret:
                    self.set_secure_cookie('user', secret)
                    self.redirect('/')
                else:
                    self.set_status(httplib.UNAUTHORIZED)
                    self.write(app._get_html('login.html'))
                    self.finish()

        return LoginHandler

    def base_routes(self):
        return [
            (r'/', self.ui_handler()),
            (r'/zounds/(.+?)/(.+?)/?', self.feature_handler()),
            (r'/zounds/login/?', self.login_handler())
        ]

    def custom_routes(self):
        return []

    def _start(self):
        tornado.ioloop.IOLoop.instance().start()

    def _secure_route(self, route):

        pattern, handler_cls = route

        if 'login' in pattern:
            return pattern, handler_cls

        class Secured(handler_cls):
            def __init__(self, *args, **kwargs):
                super(Secured, self).__init__(*args, **kwargs)

            def get_current_user(self):
                return self.get_secure_cookie('user')

            @tornado.web.authenticated
            def get(self, *args, **kwargs):
                super(Secured, self).get(*args, **kwargs)

            @tornado.web.authenticated
            def post(self, *args, **kwargs):
                super(Secured, self).post(*args, **kwargs)

            @tornado.web.authenticated
            def put(self, *args, **kwargs):
                super(Secured, self).put(*args, **kwargs)

            @tornado.web.authenticated
            def delete(self, *args, **kwargs):
                super(Secured, self).delete(*args, **kwargs)

            @tornado.web.authenticated
            def head(self, *args, **kwargs):
                super(Secured, self).head(*args, **kwargs)

            @tornado.web.authenticated
            def options(self, *args, **kwargs):
                super(Secured, self).options(*args, **kwargs)

        Secured.__name__ = handler_cls.__name__
        Secured.__module__ = handler_cls.__module__

        return pattern, Secured

    def _make_app(self):
        routes = self.custom_routes() + self.base_routes()

        if self.secret:
            routes = map(self._secure_route, routes)

        return tornado.web.Application(
            routes,
            cookie_secret=uuid4().hex,
            login_url='/zounds/login')

    def start_in_thread(self, port=8888):
        app = self._make_app()
        self.server = app.listen(port)
        self.thread = threading.Thread(target=self._start)
        self.thread.daemon = True
        self.thread.start()
        print 'Interactive REPL at http://localhost:{port}'.format(port=port)
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        ioloop = tornado.ioloop.IOLoop.instance()
        ioloop.add_callback(ioloop.stop)
        self.server.stop()
        self.thread.join()

    def start(self, port=8888):
        app = self._make_app()
        self.server = app.listen(port)
        print 'Interactive REPL at http://localhost:{port}'.format(**locals())
        self._start()



import unittest2
from sklearn_preprocessor import SklearnModel
from preprocess import PreprocessingPipeline
import featureflow as ff
import numpy as np
from zounds.util import simple_in_memory_settings
from zounds.timeseries import TimeDimension, Milliseconds, Seconds
from zounds.spectral import FrequencyBand, LinearScale, FrequencyDimension
from zounds.core import ArrayWithUnits, IdentityDimension


class MockSklearnModel(object):
    def __init__(self):
        super(MockSklearnModel, self).__init__()
        self.dims = None

    def fit(self, data):
        self.dims = data.shape[1:]
        return self

    def transform(self, data):
        return np.zeros(data.shape)

    def inverse_transform(self, data):
        return np.zeros((data.shape[0],) + self.dims)


@simple_in_memory_settings
class Document(ff.BaseModel):
    l = ff.PickleFeature(
        SklearnModel,
        model=MockSklearnModel())

    pipeline = ff.PickleFeature(
        PreprocessingPipeline,
        needs=(l,),
        store=True)


class SklearnTests(unittest2.TestCase):
    def test_should_handle_forward_transform_of_numpy_array(self):
        training_data = np.zeros((10, 5, 3))
        _id = Document.process(l=training_data)
        doc = Document(_id)
        test_data = np.zeros((11, 5, 3))
        result = doc.pipeline.transform(test_data)
        self.assertEqual((11, 15), result.data.shape)

    def test_should_handle_backward_transform_of_numpy_array(self):
        training_data = np.zeros((10, 5, 3))
        _id = Document.process(l=training_data)
        doc = Document(_id)
        test_data = np.zeros((11, 5, 3))
        result = doc.pipeline.transform(test_data)
        inverted = result.inverse_transform()
        self.assertEqual((11, 5, 3), inverted.shape)

    def test_should_preserve_time_dimension_in_forward_transform(self):
        td = TimeDimension(Seconds(1))
        td2 = TimeDimension(Milliseconds(500))
        fd = FrequencyDimension(LinearScale(FrequencyBand(10, 100), 3))
        training_data = ArrayWithUnits(
            np.zeros((10, 5, 3)), dimensions=(IdentityDimension(), td2, fd))
        _id = Document.process(l=training_data)
        doc = Document(_id)
        test_data = ArrayWithUnits(
            np.zeros((11, 5, 3)), dimensions=(td, td2, fd))
        result = doc.pipeline.transform(test_data)
        self.assertEqual((11, 15), result.data.shape)
        self.assertIsInstance(result.data, ArrayWithUnits)
        self.assertEqual(td, result.data.dimensions[0])
        self.assertEqual(IdentityDimension(), result.data.dimensions[1])

    def test_should_restore_all_dimensions_in_backward_transform(self):
        td = TimeDimension(Seconds(1))
        td2 = TimeDimension(Milliseconds(500))
        fd = FrequencyDimension(LinearScale(FrequencyBand(10, 100), 3))
        training_data = ArrayWithUnits(
            np.zeros((10, 5, 3)), dimensions=(IdentityDimension(), td2, fd))
        _id = Document.process(l=training_data)
        doc = Document(_id)
        test_data = ArrayWithUnits(
            np.zeros((11, 5, 3)), dimensions=(td, td2, fd))
        result = doc.pipeline.transform(test_data)
        inverted = result.inverse_transform()
        self.assertEqual((11, 5, 3), inverted.shape)
        self.assertIsInstance(inverted, ArrayWithUnits)
        self.assertEqual(td, inverted.dimensions[0])
        self.assertEqual(td2, inverted.dimensions[1])
        self.assertEqual(fd, inverted.dimensions[2])


import tempfile

import requests
import unittest2

from audio_metadata import AudioMetaData, MetaData
from zounds.synthesize import NoiseSynthesizer
from zounds.timeseries import SR44100, Seconds


def soundfile(flo=None):
    synth = NoiseSynthesizer(SR44100())
    samples = synth.synthesize(Seconds(5)).stereo
    flo = samples.encode(flo=flo)
    return samples, flo


class AudioMetaDataTests(unittest2.TestCase):

    def test_can_handle_local_file_path(self):
        with tempfile.NamedTemporaryFile(mode='w+') as tf:
            signal, f = soundfile(flo=tf)
            result = MetaData()._process(f.name).next()
            self.assertEqual(44100, result.samplerate)
            self.assertEqual(2, result.channels)
            self.assertIs(f.name, result.uri)

    def test_can_handle_file_like_object(self):
        samples, bio = soundfile()
        result = MetaData()._process(bio).next()
        self.assertEqual(44100, result.samplerate)
        self.assertEqual(2, result.channels)
        self.assertIs(bio, result.uri)

    def test_can_handle_url_string(self):
        url = 'http://host.com/path'
        result = MetaData()._process(url).next()
        self.assertIsInstance(result.uri, requests.Request)
        self.assertEqual(url, result.uri.url)

    def test_can_handle_request(self):
        req = requests.Request(method='GET', url='http://host.com/path')
        result = MetaData()._process(req).next()
        self.assertEqual(req, result.uri)

    def test_can_handle_audio_metadata(self):
        meta = AudioMetaData(uri='something')
        result = MetaData()._process(meta).next()
        self.assertEqual(meta, result)



import unittest2
from util import decode_timedelta
from zounds.timeseries import Seconds


class DecodeTimedeltaTests(unittest2.TestCase):
    def test_already_decoded_instance_is_returned(self):
        td = Seconds(10)
        decoded = decode_timedelta(td)
        self.assertEqual(td, decoded)



from trainer import Trainer
import numpy as np
import warnings


class SupervisedTrainer(Trainer):
    def __init__(
            self,
            model,
            loss,
            optimizer,
            epochs,
            batch_size,
            holdout_percent=0.0,
            data_preprocessor=lambda x: x,
            label_preprocessor=lambda x: x,
            checkpoint_epochs=1):

        super(SupervisedTrainer, self).__init__(
            epochs,
            batch_size,
            checkpoint_epochs=checkpoint_epochs)

        self.label_preprocessor = label_preprocessor
        self.data_preprocessor = data_preprocessor
        self.holdout_percent = holdout_percent
        self.optimizer = optimizer(model)
        self.loss = loss
        self.model = model
        self.register_batch_complete_callback(self._log)
        self.samples = None

    def _log(self, *args, **kwargs):
        if kwargs['batch'] % 10:
            return
        msg = \
            'Epoch {epoch}, batch {batch}, train error ' \
            '{train_error}, test error {test_error}'
        print msg.format(**kwargs)

    def random_sample(self):
        if self.samples is None:
            raise RuntimeError(
                'There are no samples yet.  Has training started?')
        index = np.random.randint(0, len(self.samples))
        inp, label = self.samples[index]
        return inp, label

    def train(self, data):
        import torch
        from torch.autograd import Variable

        model = self.model.cuda()
        loss = self.loss.cuda()

        data, labels = data['data'], data['labels']

        test_size = int(self.holdout_percent * len(data))
        test_data, test_labels = data[:test_size], labels[:test_size]
        data, labels = data[test_size:], labels[test_size:]

        def batch(d, l, test=False):
            d = self.data_preprocessor(d)
            l = self.label_preprocessor(l)
            inp = torch.from_numpy(d)
            inp = inp.cuda()
            inp_v = Variable(inp, volatile=test)
            output = model(inp_v)

            labels_t = torch.from_numpy(l)
            labels_t = labels_t.cuda()
            labels_v = Variable(labels_t)

            error = loss(output, labels_v)

            if not test:
                error.backward()
                self.optimizer.step()

            self.samples = zip(inp_v, output)
            return inp_v, output, error.data[0]

        start = self._current_epoch
        stop = self._current_epoch + self.checkpoint_epochs

        for epoch in xrange(start, stop):

            if epoch >= self.epochs:
                break

            for i in xrange(0, len(data), self.batch_size):

                model.zero_grad()

                # training batch
                minibatch_slice = slice(i, i + self.batch_size)
                minibatch_data = data[minibatch_slice]
                minibatch_labels = labels[minibatch_slice]

                try:
                    inp, output, e = batch(
                        minibatch_data, minibatch_labels, test=False)
                except RuntimeError as e:
                    warnings.warn(e.message)
                    continue

                # test batch
                if test_size:
                    indices = np.random.randint(0, test_size, self.batch_size)
                    test_batch_data = test_data[indices, ...]
                    test_batch_labels = test_labels[indices, ...]

                    inp, output, te = batch(
                        test_batch_data, test_batch_labels, test=True)
                else:
                    te = 'n/a'

                self.on_batch_complete(
                    epoch=epoch,
                    batch=i,
                    train_error=te,
                    test_error=e,
                    samples=self.samples)

            self._current_epoch += 1

        return model



import unittest2
from zounds.util import simple_in_memory_settings
import featureflow as ff
from preprocess import SimHash, PreprocessingPipeline
import numpy as np
from zounds.timeseries import TimeDimension, Seconds
from zounds.spectral import GeometricScale, FrequencyDimension
from zounds.core import IdentityDimension, ArrayWithUnits


@simple_in_memory_settings
class SimHashPipeline(ff.BaseModel):
    simhash = ff.PickleFeature(
        SimHash,
        bits=128)

    pipeline = ff.PickleFeature(
        PreprocessingPipeline,
        needs=(simhash,),
        store=True)


@simple_in_memory_settings
class SimHashPipelineWithPackedBits(ff.BaseModel):
    simhash = ff.PickleFeature(
        SimHash,
        packbits=True,
        bits=1024)

    pipeline = ff.PickleFeature(
        PreprocessingPipeline,
        needs=(simhash,),
        store=True)


class SimHashTests(unittest2.TestCase):
    def test_can_compute_hash_and_pack_bits(self):
        training_data = np.random.normal(0, 1, size=(100, 9))
        _id = SimHashPipelineWithPackedBits.process(simhash=training_data)
        p = SimHashPipelineWithPackedBits(_id)
        test_data = np.random.normal(0, 1, size=(1000, 9))
        result = p.pipeline.transform(test_data).data
        self.assertEqual((1000, 16), result.shape)
        self.assertEqual(np.uint64, result.dtype)

    def test_can_compute_hash_for_1d_vectors(self):
        training_data = np.random.normal(0, 1, size=(100, 9))
        _id = SimHashPipeline.process(simhash=training_data)
        p = SimHashPipeline(_id)
        test_data = np.random.normal(0, 1, size=(1000, 9))
        result = p.pipeline.transform(test_data).data
        self.assertEqual((1000, 128), result.shape)

    def test_can_compute_hash_for_2d_vectors(self):
        training_data = np.random.normal(0, 1, size=(100, 9, 7))
        _id = SimHashPipeline.process(simhash=training_data)
        p = SimHashPipeline(_id)
        test_data = np.random.normal(0, 1, size=(1000, 9, 7))
        result = p.pipeline.transform(test_data).data
        self.assertEqual((1000, 128), result.shape)

    def test_returns_array_with_units_where_possible_1d(self):
        training_data = np.random.normal(0, 1, size=(100, 9))
        _id = SimHashPipeline.process(simhash=training_data)
        p = SimHashPipeline(_id)
        td = TimeDimension(Seconds(1))
        test_data = ArrayWithUnits(
            np.random.normal(0, 1, size=(1000, 9)),
            dimensions=(
                td,
                FrequencyDimension(GeometricScale(20, 20000, 0.1, 9)))
        )
        result = p.pipeline.transform(test_data).data
        self.assertEqual((1000, 128), result.shape)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual(result.dimensions[0], td)
        self.assertIsInstance(result.dimensions[1], IdentityDimension)

    def test_returns_array_with_units_where_possible_2d(self):
        training_data = np.random.normal(0, 1, size=(100, 9, 7))
        _id = SimHashPipeline.process(simhash=training_data)
        p = SimHashPipeline(_id)
        td1 = TimeDimension(Seconds(10))
        td2 = TimeDimension(Seconds(1))
        test_data = ArrayWithUnits(
            np.random.normal(0, 1, size=(1000, 9, 7)),
            dimensions=(
                td1,
                td2,
                FrequencyDimension(GeometricScale(20, 20000, 0.1, 7)))
        )
        result = p.pipeline.transform(test_data).data
        self.assertEqual((1000, 128), result.shape)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual(result.dimensions[0], td1)
        self.assertIsInstance(result.dimensions[1], IdentityDimension)



import argparse


class BasePartialArgumentParser(argparse.ArgumentParser):
    def __init__(self, groupname, group_description):
        super(BasePartialArgumentParser, self).__init__(add_help=False)
        self.group = self.add_argument_group(groupname, group_description)

    def add_argument(self, *args, **kwargs):
        self.group.add_argument(*args, **kwargs)


class ObjectStorageSettings(BasePartialArgumentParser):
    def __init__(self):
        super(ObjectStorageSettings, self).__init__(
            'object_storage',
            'Rackspace object storage settings for model checkpoint storage')
        self.add_argument(
            '--object-storage-region',
            help='the rackspace object storage region',
            default='DFW')
        self.add_argument(
            '--object-storage-username',
            help='rackspace cloud username',
            required=True)
        self.add_argument(
            '--object-storage-api-key',
            help='rackspace cloud api key',
            required=True)


class AppSettings(BasePartialArgumentParser):
    def __init__(self):
        super(AppSettings, self).__init__(
            'app',
            'In-browser REPL settings')
        self.add_argument(
            '--app-secret',
            help='app password. If not provided, REPL is public',
            required=False)
        self.add_argument(
            '--port',
            help='The port on which the In-Browser REPL app should listen',
            default=8888)


class NeuralNetworkTrainingSettings(BasePartialArgumentParser):
    def __init__(self):
        super(NeuralNetworkTrainingSettings, self).__init__(
            'training',
            'Common settings for training neural networks')
        self.add_argument(
            '--epochs',
            help='how many passes over the data should be made during training',
            type=int)
        self.add_argument(
            '--batch-size',
            help='how many examples constitute a minibatch?',
            type=int,
            default=64)
        self.add_argument(
            '--nsamples',
            help='the number of samples to draw from the database for training',
            type=int)



import unittest2
from functional import resample
from zounds import AudioSamples
from zounds import SR11025
from zounds import SR44100
from zounds import Seconds
from zounds import SineSynthesizer, SilenceSynthesizer


class ResampleTests(unittest2.TestCase):
    def test_can_resample_audio(self):
        samplerate = SR44100()
        synth = SineSynthesizer(samplerate)
        samples = synth.synthesize(Seconds(1), [440, 880, 1760])
        encoded = samples.encode()
        new_samples = AudioSamples.from_file(encoded)
        resampled = resample(new_samples, SR11025())
        self.assertIsInstance(resampled, AudioSamples)
        self.assertEqual(int(SR11025()), len(resampled))

    def test_can_resample_stereo(self):
        samplerate = SR44100()
        synth = SineSynthesizer(samplerate)
        samples = synth.synthesize(Seconds(1), [440, 880, 1760])
        encoded = samples.stereo.encode()
        new_samples = AudioSamples.from_file(encoded)
        resampled = resample(new_samples, SR11025())
        self.assertIsInstance(resampled, AudioSamples)
        self.assertEqual(int(SR11025()), len(resampled))

    def test_resample_does_not_introduce_pops(self):
        samplerate = SR44100()
        synth = SilenceSynthesizer(samplerate)
        samples = synth.synthesize(Seconds(1))
        resampled = resample(samples, SR11025())
        self.assertEqual(0, resampled.max())


from zounds.spectral import dct_basis
import torch
from torch.autograd import Variable


class DctTransform(object):
    def __init__(self, use_cuda=False):
        super(DctTransform, self).__init__()
        self.use_cuda = use_cuda
        self._basis_cache = dict()
        self._window_cache = dict()

    def _variable(self, x, *args, **kwargs):
        v = Variable(x, *args, **kwargs)
        if self.use_cuda:
            v = v.cuda()
        return v

    def dct_basis(self, n):
        try:
            return self._basis_cache[n]
        except KeyError:
            basis = torch.from_numpy(dct_basis(n)).float()
            if self.use_cuda:
                basis = basis.cuda()
            self._basis_cache[n] = basis
            return basis

    def window(self, n, window):
        try:
            return self._window_cache[n]
        except KeyError:
            data = torch.from_numpy(window._wdata(n)).float()
            if self.use_cuda:
                data = data.cuda()
            self._window_cache[n] = data
            return data

    def _base_dct_transform(self, x, basis, axis=-1):
        n = torch.FloatTensor(1)
        n[:] = 2. / x.shape[axis]
        n = self._variable(n)
        coeffs = torch.matmul(x, basis) * torch.sqrt(n)
        return coeffs

    def dct(self, x, axis=-1):
        basis = self._variable(self.dct_basis(x.shape[axis]))
        return self._base_dct_transform(x, basis, axis)

    def idct(self, x, axis=-1):
        basis = self._variable(self.dct_basis(x.shape[axis]))
        return self._base_dct_transform(x, basis.t(), axis)

    def dct_resample(self, x, factor, axis=-1):

        # figure out how many samples our resampled signal will have
        n_samples = int(factor * x.shape[axis])

        coeffs = self.dct(x)

        # create the shape of our new coefficients
        new_coeffs_shape = list(coeffs.shape)
        new_coeffs_shape[axis] = n_samples

        # fill in the new, resampled coefficients
        new_coeffs = self._variable(torch.zeros(*new_coeffs_shape))

        new_coeffs_slices = [slice(None)] * x.dim()
        new_coeffs_slices[axis] = slice(None, coeffs.shape[axis])
        new_coeffs_slices = tuple(new_coeffs_slices)

        old_coeffs_slices = [slice(None)] * x.dim()
        old_coeffs_slices[axis] = slice(None, new_coeffs.shape[axis])
        old_coeffs_slices = tuple(old_coeffs_slices)

        new_coeffs[new_coeffs_slices] = coeffs[old_coeffs_slices]

        return self.idct(new_coeffs)

    def frequency_decomposition(self, x, factors, axis=-1):
        bands = []
        factors = sorted(factors)
        for f in factors:
            if f == 1:
                bands.append(x)
            else:
                rs = self.dct_resample(x, f, axis)
                bands.append(rs)
                us = self.dct_resample(rs, (1. / f), axis)
                x = x - us
        return bands

    def short_time_dct(self, x, size, step, window):
        original_shape = x.shape
        x = x.unfold(-1, size, step)
        window = self._variable(self.window(x.shape[-1], window))
        x = x * window
        x = self.dct(x, axis=-1)
        x = x.view((original_shape[0], size, x.shape[2]))
        return x



from featureflow import Node, NotEnoughData
import marshal
import types
from collections import OrderedDict
import hashlib
import numpy as np
from functional import hyperplanes
from zounds.loudness import log_modulus, inverse_log_modulus


class Op(object):
    def __init__(self, func, **kwargs):
        super(Op, self).__init__()
        self._kwargs = kwargs
        try:
            self._func = marshal.dumps(func.func_code)
        except AttributeError:
            # func is already a marshalled function
            self._func = func
        self._version = self._compute_version()

    @property
    def version(self):
        return self._version

    @property
    def kwargs(self):
        return self._kwargs

    def _compute_version(self):
        h = hashlib.md5(self._func)
        for v in self._kwargs.itervalues():
            try:
                h.update(v.version)
            except AttributeError:
                h.update(np.array(v))
        return h.hexdigest()

    def __getattr__(self, key):
        if key == '_kwargs':
            raise AttributeError()

        try:
            return self._kwargs[key]
        except KeyError:
            raise AttributeError(key)

    def __call__(self, arg, **kwargs):
        code = marshal.loads(self._func)
        f = types.FunctionType(
            code,
            globals(),
            'preprocess')
        kwargs.update(self._kwargs)
        return f(arg, **kwargs)


class PreprocessResult(object):
    """
    `PreprocessResult` are the output of :class:`Preprocessor` nodes, and can
    participate in a `Pipeline`.

    Args:
        data: the data on which the node in the graph was originally trained
        op (Op): a callable that can transform data
        inversion_data: data extracted in the forward pass of the model, that
            can be used to invert the result
        inverse (Op): a callable that given the output of `op`, and
            `inversion_data`, can invert the result
    """

    def __init__(self, data, op, inversion_data=None, inverse=None, name=None):
        super(PreprocessResult, self).__init__()
        self.name = name
        self.inverse = inverse
        self.inversion_data = inversion_data
        self.data = data
        self.op = op

    def __str__(self):
        return 'PreprocessResult(name={name})'.format(**self.__dict__)

    def __repr__(self):
        return self.__str__()

    def __getattr__(self, key):
        if key == 'op':
            raise AttributeError()
        return getattr(self.op, key)

    def for_storage(self):
        return PreprocessResult(
            None,
            self.op,
            self.inversion_data,
            self.inverse,
            self.name)


class Preprocessor(Node):
    """
    `Preprocessor` is the common base class for nodes in a processing graph that
    will produce :class:`PreprocessingResult` instances that end up as part of
    a :class:`Pipeline`.

    Args:
        needs (Node): previous processing node(s) on which this one depends
            for its data

    See Also:
        :class:`PreprocessResult`
        :class:`PreprocessingPipeline`
        :class:`PipelineResult`
    """

    def __init__(self, needs=None):
        super(Preprocessor, self).__init__(needs=needs)

    def _forward_func(self):
        """
        Return a function that represents this processor's
        forward transform
        """

        def x(data, example_arg=None, another_one=None):
            return data

        return x

    def transform(self, **kwargs):
        return Op(self._forward_func(), **kwargs)

    def _inversion_data(self):
        """
        Return a function that computes any data needed for
        an inverse transform, if one is possible.  Otherwise,
        raise NotImplemented
        """

        def x(data, **kwargs):
            return kwargs

        return x

    def inversion_data(self, **kwargs):
        return Op(self._inversion_data(), **kwargs)

    def _backward_func(self):
        """
        Return a function that computes this processor's
        inverse transform, if one is possible.  Otherwise,
        raise NotImplemented
        """

        def x(data, **inversion_args):
            return data

        return x

    def inverse_transform(self):
        return Op(self._backward_func())

    def _extract_data(self, data):
        if isinstance(data, PreprocessResult):
            return data.data
        elif isinstance(data, dict):
            return dict((k, self._extract_data(v)) for k, v in data.iteritems())
        else:
            return data


class UnitNorm(Preprocessor):
    def __init__(self, needs=None):
        super(UnitNorm, self).__init__(needs=needs)

    def _forward_func(self):
        def x(d):
            from zounds.core import ArrayWithUnits
            from functional import example_wise_unit_norm
            normed, norms = example_wise_unit_norm(d, return_norms=True)
            try:
                normed = ArrayWithUnits(normed, d.dimensions)
            except AttributeError:
                pass

            return ForwardResult(normed, norm=norms)

        return x

    def _backward_func(self):
        def x(d, norm=None):
            dim_diff = d.ndim - norm.ndim
            new_dims = (1,) * dim_diff
            return d * norm.reshape(norm.shape + new_dims)

        return x

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform()
        inv_data = self.inversion_data()
        inv = self.inverse_transform()
        try:
            forward_result = op(data)
            data = forward_result.output_value
        except AttributeError:
            data = None
        yield PreprocessResult(
            data, op, inversion_data=inv_data, inverse=inv, name='UnitNorm')


class BasePreprocessor(Preprocessor):
    def __init__(self, forward=None, backward=None, name=None, needs=None):
        super(BasePreprocessor, self).__init__(needs=needs)

        def not_implemented():
            raise NotImplementedError()

        self.forward = forward
        self.backward = backward or not_implemented
        self.name = name

    def _forward_func(self):
        return self.forward

    def _backward_func(self):
        return self.backward

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform()
        inv_data = self.inversion_data()
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
            data, op, inversion_data=inv_data, inverse=inv, name=self.name)


# TODO: what about functions with imports that aren't in the calling namespace?
# TODO: what about functions that require inversion data?
# TODO: what about functions that need to do some processing first (e.g. SimHash)
class Log(BasePreprocessor):
    def __init__(self, needs=None):
        super(Log, self).__init__(
            forward=log_modulus,
            backward=inverse_log_modulus,
            name='Log',
            needs=needs)


class MuLawCompressed(Preprocessor):
    def __init__(self, needs=None):
        super(MuLawCompressed, self).__init__(needs=needs)

    def _forward_func(self):
        def x(d):
            from zounds.loudness import mu_law
            return mu_law(d)

        return x

    def _backward_func(self):
        def x(d):
            from zounds.loudness import inverse_mu_law
            return inverse_mu_law(d)

        return x

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform()
        inv_data = self.inversion_data()
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
            data,
            op,
            inversion_data=inv_data,
            inverse=inv,
            name='MuLawCompressed')


class Slicer(Preprocessor):
    def __init__(self, slicex=None, needs=None):
        super(Slicer, self).__init__(needs=needs)
        self.fill_func = np.zeros
        self.slicex = slicex

    def _forward_func(self):
        def x(d, slicex=None):
            return d[..., slicex]

        return x

    def _inversion_data(self):
        def y(d, slicex=None, fill_func=None):

            try:
                ka = d.kwargs()

                def ff(shape, dtype):
                    return d.__class__(fill_func(shape, dtype), **ka)
            except AttributeError:

                def ff(shape, dtype):
                    return fill_func(shape, dtype)

            return dict(shape=d.shape, slicex=slicex, fill_func=ff)

        return y

    def _backward_func(self):
        def z(d, shape=None, fill_func=None, slicex=None):
            new_shape = d.shape[:1] + shape[1:]
            new_arr = fill_func(new_shape, d.dtype)
            new_arr[..., slicex] = d
            return new_arr

        return z

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform(slicex=self.slicex)
        inv_data = self.inversion_data(
            fill_func=self.fill_func, slicex=self.slicex)
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
            data, op, inversion_data=inv_data, inverse=inv, name='Slicer')


class Reshape(Preprocessor):
    def __init__(self, new_shape, needs=None):
        super(Reshape, self).__init__(needs=needs)
        self.new_shape = new_shape

    def _forward_func(self):
        def x(d, new_shape=None):
            return d.reshape(d.shape[:1] + new_shape)

        return x

    def _inversion_data(self):
        def x(d):
            return dict(original_shape=d.shape)

        return x

    def _backward_func(self):
        def x(d, original_shape=None):
            return d.reshape(original_shape)

        return x

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform(new_shape=self.new_shape)
        inv_data = self.inversion_data()
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
            data, op, inversion_data=inv_data, inverse=inv, name='Reshape')


class SimHash(Preprocessor):
    """
    Hash feature vectors by computing on which side of N hyperplanes those
    features lie.

    Args:
        bits (int): The number of hyperplanes, and hence, the number of bits
            in the resulting hash
        packbits (bool): Should the result be bit-packed?
        needs (Preprocessor): the processing node on which this node relies for
            its data
    """

    def __init__(self, bits=None, packbits=False, needs=None):
        super(SimHash, self).__init__(needs=needs)
        self.packbits = packbits
        self.bits = bits

    def _forward_func(self):
        def x(d, plane_vectors=None, packbits=None):

            from zounds.core import ArrayWithUnits, IdentityDimension
            from zounds.learn import simhash
            import numpy as np

            bits = simhash(plane_vectors, d)

            if packbits:
                bits = np.packbits(bits, axis=-1).view(np.uint64)

            try:
                return ArrayWithUnits(
                    bits, [d.dimensions[0], IdentityDimension()])
            except AttributeError:
                return bits

        return x

    def _backward_func(self):
        def x(d):
            raise NotImplementedError()

        return x

    def _process(self, data):
        data = self._extract_data(data)
        mean = data.mean(axis=0).flatten()
        std = data.std(axis=0).flatten()
        plane_vectors = hyperplanes(mean, std, self.bits)
        op = self.transform(plane_vectors=plane_vectors, packbits=self.packbits)
        inv_data = self.inversion_data()
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
            data, op, inversion_data=inv_data, inverse=inv, name='SimHash')


class MeanStdNormalization(Preprocessor):
    def __init__(self, needs=None):
        super(MeanStdNormalization, self).__init__(needs=needs)

    def _forward_func(self):
        def x(d, mean=None, std=None):
            import numpy as np
            return np.divide(d - mean, std, where=std != 0)

        return x

    def _backward_func(self):
        def x(d, mean=None, std=None):
            arr = d.copy()
            arr *= std
            arr += mean
            return arr

        return x

    def _process(self, data):
        data = self._extract_data(data)
        mean = data.mean(axis=0)
        std = data.std(axis=0)
        op = self.transform(mean=mean, std=std)
        inv_data = self.inversion_data(mean=mean, std=std)
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
            data, op, inversion_data=inv_data, inverse=inv, name='MeanStd')


class ForwardResult(object):
    def __init__(self, output_value, **inversion_data):
        self.output_value = output_value
        self.inversion_data = inversion_data


class InstanceScaling(Preprocessor):
    def __init__(self, max_value=1, needs=None):
        super(InstanceScaling, self).__init__(needs=needs)
        self.max_value = max_value

    def _forward_func(self):
        def x(d, max_value=None):
            import numpy as np
            axes = tuple(range(1, len(d.shape)))
            # TODO: it should be possible to hand this intermediate value off
            # to inversion data, and to ignore the extra
            m = np.max(np.abs(d), axis=axes, keepdims=True)
            output = max_value * np.divide(d, m, where=m != 0)
            return ForwardResult(output, max=m, max_value=max_value)

        return x

    def _backward_func(self):
        def x(d, max=None, max_value=None):
            return (d * max) / max_value

        return x

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform(max_value=self.max_value)
        inv_data = self.inversion_data()
        inv = self.inverse_transform()
        forward_result = op(data)
        data = forward_result.output_value
        yield PreprocessResult(
            data,
            op,
            inversion_data=inv_data,
            inverse=inv,
            name='InstanceScaling')


class Multiply(Preprocessor):
    def __init__(self, factor=1, needs=None):
        super(Multiply, self).__init__(needs=needs)
        self.factor = factor

    def _forward_func(self):
        def x(d, factor=None):
            return d * factor

        return x

    def _backward_func(self):
        def x(d, factor=None):
            return d * (1.0 / factor)

        return x

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform(factor=self.factor)
        inv_data = self.inversion_data(factor=self.factor)
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
            data, op, inversion_data=inv_data, inverse=inv, name='Multiply')


class Weighted(Preprocessor):
    def __init__(self, weighting, needs=None):
        super(Weighted, self).__init__(needs=needs)
        self.weighting = weighting

    def _forward_func(self):
        def x(d, weighting=None):
            return d * weighting

        return x

    def _backward_func(self):
        def x(d, weighting=None):
            return d / weighting

        return x

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform(weighting=self.weighting)
        inv_data = self.inversion_data(weighting=self.weighting)
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
            data, op, inversion_data=inv_data, inverse=inv, name='Weighted')


class AbsoluteValue(Preprocessor):
    def __init__(self, needs=None):
        super(AbsoluteValue, self).__init__(needs=needs)

    def _forward_func(self):
        def x(d):
            import numpy as np
            from zounds.core import ArrayWithUnits
            processed = np.abs(d)
            try:
                return ArrayWithUnits(processed, d.dimensions)
            except AttributeError:
                return processed

        return x

    def _backward_func(self):
        def x(d):
            raise NotImplementedError()

        return x

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform()
        inv_data = self.inversion_data()
        inv = self.inverse_transform()

        try:
            data = op(data)
        except:
            data = None

        yield PreprocessResult(
            data, op, inversion_data=inv_data, inverse=inv,
            name='AbsoluteValue')


class Sharpen(Preprocessor):
    def __init__(self, kernel=None, needs=None):
        super(Sharpen, self).__init__(needs=needs)
        self.kernel = kernel

    def _forward_func(self):
        def x(d, kernel=None):
            from scipy.signal import convolve
            from zounds.core import ArrayWithUnits
            data = convolve(d, kernel[None, ...], mode='same')
            data[data < 0] = 0
            try:
                return ArrayWithUnits(data, d.dimensions)
            except AttributeError:
                return data

        return x

    def _backward_func(self):
        def x(d):
            raise NotImplementedError()

        return x

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform(kernel=self.kernel)
        inv_data = self.inversion_data()
        inv = self.inverse_transform()
        try:
            data = op(data)
        except:
            data = None
        yield PreprocessResult(
            data, op, inversion_data=inv_data, inverse=inv, name='Sharpen')


class Binarize(Preprocessor):
    def __init__(self, threshold=0.5, needs=None):
        super(Binarize, self).__init__(needs=needs)
        self.threshold = threshold

    def _forward_func(self):
        def x(d, threshold=None):
            import numpy as np
            from zounds.core import ArrayWithUnits
            data = np.zeros(d.shape, dtype=np.uint8)
            data[np.where(d > threshold)] = 1
            try:
                return ArrayWithUnits(data, d.dimensions)
            except AttributeError:
                return data

        return x

    def _backward_func(self):
        def x(d):
            raise NotImplementedError()

        return x

    def _process(self, data):
        data = self._extract_data(data)
        op = self.transform(threshold=self.threshold)
        inv_data = self.inversion_data()
        inv = self.inverse_transform()
        try:
            data = op(data)
        except:
            data = None
        yield PreprocessResult(
            data, op, inversion_data=inv_data, inverse=inv, name='Binarize')


class Pipeline(object):
    """

    """

    def __init__(self, preprocess_results):
        self.processors = list(preprocess_results)
        self.version = hashlib.md5(
            ''.join([p.op.version for p in self.processors])).hexdigest()

    def __getitem__(self, index):
        if isinstance(index, int):
            return self.processors[index]
        elif isinstance(index, list):
            return Pipeline([self.processors[i] for i in index])
        return Pipeline(self.processors[index])

    def wrap_data(self, data):
        cls = data.__class__
        try:
            kwargs = data.kwargs()
        except AttributeError:
            kwargs = None
        return cls, kwargs

    def transform(self, data, wrapper=None):
        inversion_data = []
        wrap_data = []
        for p in self.processors:
            wrap_data.append(self.wrap_data(data))
            processed_data = p.op(data)

            if isinstance(processed_data, ForwardResult):
                inv_data = processed_data.inversion_data
                processed_data = processed_data.output_value
            else:
                inv_data = p.inversion_data(data)

            inversion_data.append(inv_data)
            data = processed_data

        if wrapper is not None:
            data = wrapper(data)

        return PipelineResult(data, self.processors, inversion_data, wrap_data)


class PipelineResult(object):
    def __init__(self, data, processors, inversion_data, wrap_data):
        super(PipelineResult, self).__init__()
        self.processors = processors[::-1]
        self.inversion_data = inversion_data[::-1]
        self.wrap_data = wrap_data[::-1]
        self.data = data

    def unwrap(self, data, wrap_data):
        cls, kwargs = wrap_data
        if kwargs is None:
            return data
        return cls(data, **kwargs)

    def inverse_transform(self):
        data = self.data
        for inv_data, wrap_data, p in \
                zip(self.inversion_data, self.wrap_data, self.processors):
            data = p.inverse(data, **inv_data)
            data = self.unwrap(data, wrap_data)
        return data


class PreprocessingPipeline(Node):
    """
    A `PreprocessingPipeline` is a node in the graph that can be connected to
    one or more :class:`Preprocessor` nodes, whose output it will assemble into
    a re-usable pipeline.

    Args:
        needs (list or tuple of Node): the :class:`Preprocessor` nodes on whose
            output this pipeline depends

    Here's an example of a learning pipeline that will first find the
    feature-wise mean and standard deviation of a dataset, and will then learn
    K-Means clusters from the dataset.  This will result in a re-usable pipeline
    that can use statistics from the original dataset to normalize new examples,
    assign them to a cluster, and finally, reconstruct them.

    .. code:: python

        import featureflow as ff
        import zounds
        from random import choice

        samplerate = zounds.SR44100()
        STFT = zounds.stft(resample_to=samplerate)


        @zounds.simple_in_memory_settings
        class Sound(STFT):
            bark = zounds.ArrayWithUnitsFeature(
                zounds.BarkBands,
                samplerate=samplerate,
                needs=STFT.fft,
                store=True)


        @zounds.simple_in_memory_settings
        class ExamplePipeline(ff.BaseModel):
            docs = ff.PickleFeature(
                ff.IteratorNode,
                needs=None)

            shuffled = ff.PickleFeature(
                zounds.ShuffledSamples,
                nsamples=100,
                needs=docs,
                store=False)

            meanstd = ff.PickleFeature(
                zounds.MeanStdNormalization,
                needs=docs,
                store=False)

            kmeans = ff.PickleFeature(
                zounds.KMeans,
                needs=meanstd,
                centroids=32)

            pipeline = ff.PickleFeature(
                zounds.PreprocessingPipeline,
                needs=(meanstd, kmeans),
                store=True)

        # apply the Sound processing graph to individual audio files
        for metadata in zounds.InternetArchive('TheR.H.SFXLibrary'):
            print 'processing {url}'.format(url=metadata.request.url)
            Sound.process(meta=metadata)

        # apply the ExamplePipeline processing graph to the entire corpus of audio
        _id = ExamplePipeline.process(docs=(snd.bark for snd in Sound))
        learned = ExamplePipeline(_id)

        snd = choice(list(Sound))
        result = learned.pipeline.transform(snd.bark)
        print result.data  # print the assigned centroids for each FFT frame
        inverted = result.inverse_transform()
        print inverted  # the reconstructed FFT frames


    See Also:
        :class:`Pipeline`
        :class:`Preprocessor`
        :class:`PreprocessResult`
        :class:`PipelineResult`
    """

    def __init__(self, needs=None):
        super(PreprocessingPipeline, self).__init__(needs=needs)
        self._init_pipeline()

    def _init_pipeline(self):
        self._pipeline = OrderedDict((id(n), None) for n in self.needs.values())

    def _enqueue(self, data, pusher):
        self._pipeline[id(pusher)] = data

    def _dequeue(self):
        if not all(self._pipeline.itervalues()):
            raise NotEnoughData()

        pipeline = Pipeline(map(
            lambda x: x.for_storage(),
            self._pipeline.itervalues()))

        self._init_pipeline()
        return pipeline



import unittest2

try:
    from dct_transform import DctTransform
    import torch
    from torch.autograd import Variable
except ImportError:
    DctTransform = None


class DctTransformTests(unittest2.TestCase):
    def setUp(self):
        if DctTransform is None:
            self.skipTest('PyTorch is not available')

    def test_can_do_short_time_dct_transform(self):
        t = torch.FloatTensor(3, 1, 512)
        v = Variable(t)
        dct_trasform = DctTransform()
        stdct = dct_trasform.short_time_dct(v, 128, 64)
        self.assertEqual((3, 128, 7), tuple(stdct.shape))



import requests
import time
from zounds.soundfile import AudioMetaData


class FreeSoundSearch(object):
    """
    Produces an iterable of :class:`zounds.soundfile.AudioMetaData` instances
    for every result from a https://freesound.org search

    Args:
        api_key (str): Your freesound.org API key (get one here:
            (http://freesound.org/apiv2/apply/))
        query (str): The text query to perform

    Raises
        ValueError: when `api_key` and/or `query` are not supplied

    Examples:
        >>> from zounds import FreeSoundSearch
        >>> fss = FreeSoundSearch('YOUR_API_KEY', 'guitar')
        >>> iter(fss).next()
        {'description': u'Etude of Electric Guitar in Dm. Used chorus and reverberation effects. Size 6/4. Tempo 100. Gloomy and sentimental.', 'tags': [u'Etude', u'Experemental', u'Guitar', u'guitar', u'Electric', u'Chorus'], 'uri': <Request [GET]>, 'channels': 2, 'licensing': u'http://creativecommons.org/licenses/by/3.0/', 'samplerate': 44100.0}

    See Also:
        :class:`InternetArchive`
        :class:`PhatDrumLoops`
        :class:`zounds.soundfile.AudioMetaData`
    """

    def __init__(self, api_key, query, n_results=10, delay=0.2):
        super(FreeSoundSearch, self).__init__()

        if not api_key:
            raise ValueError('You must supply a freesound.org API key')

        if not query:
            raise ValueError('You must supply a text query')

        self.delay = delay
        self.n_results = n_results
        self.query = query
        self.api_key = api_key

    def _get_metadata_by_id(self, _id):
        sound_data = requests.get(
            'http://www.freesound.org/apiv2/sounds/{_id}'.format(**locals()),
            params={'token': self.api_key}
        )
        sound_data.raise_for_status()
        return sound_data.json()

    def _freesound_to_audio_metadata(self, data):
        request = requests.Request(
            method='GET',
            url=data['previews']['preview-hq-ogg'],
            params={'token': self.api_key})

        web_url = 'https://freesound.org/people/{username}/sounds/{id}/'\
            .format(**data)

        return AudioMetaData(
            uri=request,
            samplerate=data['samplerate'],
            channels=data['channels'],
            licensing=data['license'],
            description=data['description'],
            tags=data['tags'],
            web_url=web_url)

    def get_by_id(self, freesound_id):
        data = self._get_metadata_by_id(freesound_id)
        print data
        return self._freesound_to_audio_metadata(data)

    def _iter_results(self, link=None):
        if link:
            results = requests.get(link, params={'token': self.api_key})
        else:
            results = requests.get(
                'http://www.freesound.org/apiv2/search/text',
                params={
                    'query': self.query,
                    'token': self.api_key
                })

        results.raise_for_status()
        results = results.json()

        for r in results['results']:
            yield self._get_metadata_by_id(r['id'])

            # prevent 429 "Too Many Requests" responses
            time.sleep(0.2)

        for r in self._iter_results(results['next']):
            yield r

    def __iter__(self):
        for i, data in enumerate(self._iter_results()):
            if i > self.n_results:
                break
            yield self._freesound_to_audio_metadata(data)



from __future__ import division
import numpy as np
from featureflow import Node


def log_modulus(x):
    return np.sign(x) * np.log(np.abs(x) + 1)


def inverse_log_modulus(x):
    return (np.exp(np.abs(x)) - 1) * np.sign(x)


def decibel(x):
    return 20 * np.log10(x)


def mu_law(x, mu=255):
    s = np.sign(x)
    x = np.abs(x)
    return s * (np.log(1 + (mu * x)) / np.log(1 + mu))


def inverse_mu_law(x, mu=255):
    s = np.sign(x)
    x = np.abs(x)
    x *= np.log(1 + mu)
    x = (np.exp(x) - 1) / mu
    return x * s


def unit_scale(x, axis=None):
    scaled = x - x.min(axis=axis, keepdims=True)
    mx = scaled.max(axis=axis, keepdims=True)
    scaled = np.divide(scaled, mx, where=max != 0)
    return scaled


class MuLaw(Node):
    def __init__(self, mu=255, needs=None):
        super(MuLaw, self).__init__(needs=needs)
        self.mu = mu

    def _process(self, data):
        yield mu_law(data, mu=self.mu)


class LogModulus(Node):
    def __init__(self, factor=1, needs=None):
        super(LogModulus, self).__init__(needs=needs)
        self.factor = factor

    def _process(self, data):
        yield log_modulus(data * self.factor)



from trainer import Trainer
from random import choice
import numpy as np


class TripletEmbeddingTrainer(Trainer):
    """
    Learn an embedding by applying the triplet loss to anchor examples, negative
    examples, and deformed or adjacent examples, akin to:

    * `UNSUPERVISED LEARNING OF SEMANTIC AUDIO REPRESENTATIONS` <https://arxiv.org/pdf/1711.02209.pdf>

    Args:
        network (nn.Module): the neural network to train
        epochs (int): the desired number of passes over the entire dataset
        batch_size (int): the number of examples in each minibatch
        anchor_slice (slice): since choosing examples near the anchor example
            is one possible transformation that can be applied to find a positive
            example, batches generally consist of examples that are longer
            (temporally) than the examples that will be fed to the network, so
            that adjacent examples may be chosen.  This slice indicates which
            part of the minibatch examples comprises the anchor
        deformations (callable): a collection of other deformations or
            transformations that can be applied to anchor examples to derive
            positive examples.  These callables should take two arguments: the
            anchor examples from the minibatch, as well as the "wider" minibatch
            examples that include temporally adjacent events
    """

    def __init__(
            self,
            network,
            epochs,
            batch_size,
            anchor_slice,
            deformations=None,
            checkpoint_epochs=1):

        super(TripletEmbeddingTrainer, self).__init__(
            epochs, batch_size, checkpoint_epochs=checkpoint_epochs)
        self.anchor_slice = anchor_slice
        self.network = network
        self.deformations = deformations

        # The margin hyperparameter is set to 0.1 in, according to section 4.2
        # of the paper https://arxiv.org/pdf/1711.02209.pdf
        self.margin = 0.1
        self.register_batch_complete_callback(self._log)

    def _driver(self, data):
        batches_in_epoch = len(data) // self.batch_size
        start = self._current_epoch
        stop = self._current_epoch + self.checkpoint_epochs
        for epoch in xrange(start, stop):
            if epoch > self.epochs:
                break

            for batch in xrange(batches_in_epoch):
                yield epoch, batch

            self._current_epoch += 1

    def _apply_network_and_normalize(self, x):
        """
        Pass x through the network, and give the output unit norm, as specified
        by section 4.2 of https://arxiv.org/pdf/1711.02209.pdf
        """

        import torch
        x = self.network(x)
        return x / torch.norm(x, dim=1).view(-1, 1)

    def _select_batch(self, training_set):
        indices = np.random.randint(0, len(training_set), self.batch_size)
        batch = training_set[indices, self.anchor_slice]
        return indices, batch.astype(np.float32)

    def train(self, data):
        from torch import nn
        from torch.optim import Adam
        from util import to_var

        data = data['data']

        self.network.cuda()
        self.network.train()

        optimizer = Adam(self.network.parameters(), lr=1e-5)
        loss = nn.TripletMarginLoss(margin=self.margin).cuda()

        for epoch, batch in self._driver(data):
            self.network.zero_grad()

            # choose a batch of anchors
            indices, anchor = self._select_batch(data)
            anchor_v = to_var(anchor)
            a = self._apply_network_and_normalize(anchor_v)

            # choose negative examples
            negative_indices, negative = self._select_batch(data)
            negative_v = to_var(negative)
            n = self._apply_network_and_normalize(negative_v)

            # choose a deformation for this batch and apply it to produce the
            # positive examples
            deformation = choice(self.deformations)
            positive = deformation(anchor, data[indices, ...]) \
                .astype(np.float32)
            positive_v = to_var(positive)
            p = self._apply_network_and_normalize(positive_v)

            error = loss.forward(a, p, n)
            error.backward()
            optimizer.step()

            self.on_batch_complete(
                epoch=epoch,
                batch=batch,
                error=float(error.data.cpu().numpy().squeeze()),
                deformation=deformation.__name__)

        return self.network

    def _log(self, *args, **kwargs):
        print \
            'epoch {epoch}, batch {batch}, ' \
            'error {error}, deformation {deformation}'.format(**kwargs)



from featureflow import Node, NotEnoughData
from zounds.core import ArrayWithUnits, IdentityDimension
import numpy as np
from multiprocessing.pool import ThreadPool, cpu_count


class Reservoir(object):
    def __init__(self, nsamples, dtype=None):
        super(Reservoir, self).__init__()

        if not isinstance(nsamples, int):
            raise ValueError('nsamples must be an integer')

        if nsamples <= 0:
            raise ValueError('nsamples must be greater than zero')

        self.nsamples = nsamples
        self.arr = None
        self.indices = set()
        self.dtype = dtype

    def percent_full(self):
        return float(len(self.indices)) / self.nsamples

    def _init_arr(self, samples):
        if self.arr is not None:
            return

        shape = (self.nsamples,) + samples.shape[1:]
        self.arr = np.zeros(shape, dtype=self.dtype or samples.dtype)

        try:
            self.arr = ArrayWithUnits(
                self.arr, (IdentityDimension(),) + samples.dimensions[1:])
        except AttributeError:
            pass

    def add(self, samples, indices=None):
        self._init_arr(samples)

        if indices is None:
            indices = np.random.randint(0, self.nsamples, len(samples))

        if len(indices) != len(samples):
            raise ValueError(
                'number of input samples and indices must match'
                ' but they were {samples} and {indices} respectively'
                    .format(samples=len(samples), indices=len(indices)))

        self.arr[indices, ...] = samples
        self.indices.update(indices)

    def get(self):
        if len(self.indices) == self.nsamples:
            return self.arr

        return self.arr[sorted(self.indices), ...]

    def get_batch(self, batch_size):
        if batch_size > self.nsamples:
            raise ValueError(
                'Requested {batch_size} samples, but this instance can provide '
                'at maximum {nsamples}'
                    .format(batch_size=batch_size, nsamples=self.nsamples))

        if batch_size > len(self.indices):
            raise ValueError(
                'Requested {batch_size} samples, but this instance only '
                'currently has {n} samples, with a maximum of {nsamples}'
                    .format(
                    batch_size=batch_size,
                    n=len(self.indices),
                    nsamples=self.nsamples))

        # TODO: this would be much more efficient for repeated calls if I
        # instead maintained a sorted set
        indices = np.random.choice(list(self.indices), batch_size)
        return self.arr[indices, ...]


class MultiplexedReservoir(object):
    def __init__(self, nsamples, dtype=None):
        super(MultiplexedReservoir, self).__init__()
        self.dtype = dtype
        self.reservoir = None
        self.nsamples = nsamples

    def _init_dict(self, samples):
        if self.reservoir is not None:
            return

        if self.reservoir is None:
            self.reservoir = dict(
                (k, Reservoir(self.nsamples, dtype=self.dtype))
                for k in samples.iterkeys())

    def _check_sample_keys(self, samples):
        if set(self.reservoir.keys()) != set(samples.keys()):
            raise ValueError(
                'samples should have keys {keys}'
                    .format(keys=self.reservoir.keys()))

    def add(self, samples):
        self._init_dict(samples)
        self._check_sample_keys(samples)

        indices = None
        for k, v in samples.iteritems():
            if indices is None:
                indices = np.random.randint(0, self.nsamples, len(v))

            self.reservoir[k].add(v, indices=indices)

    def get(self):
        return dict((k, v.get()) for k, v in self.reservoir.iteritems())


class ShuffledSamples(Node):
    def __init__(
            self,
            nsamples=None,
            multiplexed=False,
            dtype=None,
            needs=None):
        super(ShuffledSamples, self).__init__(needs=needs)
        self.reservoir = MultiplexedReservoir(nsamples, dtype=dtype) \
            if multiplexed else Reservoir(nsamples, dtype=dtype)

    def _enqueue(self, data, pusher):
        self.reservoir.add(data)

    def _dequeue(self):
        if not self._finalized:
            raise NotEnoughData()
        return self.reservoir.get()


class InfiniteSampler(Node):
    def __init__(self, nsamples=None, dtype=None, needs=None):
        super(InfiniteSampler, self).__init__(needs=needs)
        self.reservoir = Reservoir(nsamples, dtype)

    def _total_samples(self, cls, feature, _ids):
        pool = ThreadPool(cpu_count())
        total_samples = sum(pool.imap_unordered(
            lambda _id: len(feature(_id=_id, persistence=cls)), _ids))
        return total_samples

    def _update_reservoir(self, _id, cls, feature, total_samples):
        # fetch the features from a single document
        x = feature(_id=_id, persistence=cls)

        # compute the contribution this sample makes to the dataset at
        # large
        feature_size = len(x)
        ratio = float(feature_size) / total_samples

        # determine the appropriate number of samples to contribute to
        # the reservoir
        nsamples = max(1, int(self.reservoir.nsamples * ratio))

        print 'Contributing', feature_size, ratio, nsamples

        # select an appropriately-sized and random subset of the feature.
        # this will be shuffled again as it is added to the reservoir,
        # but this ensures that samples are drawn evenly from the
        # duration of the sound
        indices = np.random.randint(0, feature_size, nsamples)

        self.reservoir.add(x[indices, ...])
        return len(indices)

    def _process(self, data):
        cls, feature = data

        # compute the total number of samples in our dataset
        _ids = list(cls.database.iter_ids())
        total_samples = self._total_samples(cls, feature, _ids)
        print 'Total samples', total_samples

        while True:
            pool = ThreadPool(cpu_count())
            list(pool.imap_unordered(
                lambda _id: self._update_reservoir(
                    _id, cls, feature, total_samples),
                _ids))

            yield self.reservoir.get()


class ReservoirSampler(Node):
    """
    Use reservoir sampling (http://en.wikipedia.org/wiki/Reservoir_sampling) to
    draw a fixed-size set of random samples from a stream of unknown size.

    This is useful when the samples can fit into memory, but the stream cannot.
    """

    def __init__(self, nsamples=None, wrapper=None, needs=None):
        super(ReservoirSampler, self).__init__(needs=needs)
        if wrapper:
            raise DeprecationWarning('wrapper is no longer used or needed')
        self._nsamples = int(nsamples)
        self._r = None
        self._index = 0

    # TODO: What happens if we have filled up all the sample slots and we run
    # out of data?
    def _enqueue(self, data, pusher):
        if self._r is None:
            shape = (self._nsamples,) + data.shape[1:]
            self._r = np.zeros(shape, dtype=data.dtype)
            try:
                self._r = ArrayWithUnits(
                    self._r, (IdentityDimension(),) + data.dimensions[1:])
            except AttributeError:
                # samples were likely a plain numpy array, and not an
                # ArrayWithUnits instance
                pass

        diff = 0
        if self._index < self._nsamples:
            diff = self._nsamples - self._index
            available = len(data[:diff])
            self._r[self._index: self._index + available] = data[:diff]
            self._index += available

        remaining = len(data[diff:])
        if not remaining:
            return
        indices = np.random.random_integers(0, self._index, size=remaining)
        indices = indices[indices < self._nsamples]
        self._r[indices, ...] = data[diff:][range(len(indices))]
        self._index += remaining

    def _dequeue(self):
        if not self._finalized:
            raise NotEnoughData()

        if self._index <= self._nsamples:
            arr = np.asarray(self._r[:self._index])
            np.random.shuffle(arr)
            if isinstance(self._r, ArrayWithUnits):
                arr = ArrayWithUnits(arr, self._r.dimensions)
            return arr

        return self._r



import numpy as np
from featureflow import BaseModel, JSONFeature, ByteStream, ByteStreamFeature
from zounds.soundfile import \
    MetaData, AudioMetaDataEncoder, OggVorbis, OggVorbisFeature, AudioStream, \
    Resampler, ChunkSizeBytes
from zounds.segment import \
    ComplexDomain, MovingAveragePeakPicker, TimeSliceFeature
from zounds.persistence import ArrayWithUnitsFeature, AudioSamplesFeature, \
    FrequencyAdaptiveFeature
from zounds.timeseries import SR44100, HalfLapped, Stride, Seconds
from zounds.spectral import \
    SlidingWindow, OggVorbisWindowingFunc, FFT, BarkBands, SpectralCentroid, \
    Chroma, BFCC, DCT, FrequencyAdaptiveTransform, FrequencyBand

DEFAULT_CHUNK_SIZE = ChunkSizeBytes(
    samplerate=SR44100(),
    duration=Seconds(30),
    bit_depth=16,
    channels=2)


def resampled(
        chunksize_bytes=DEFAULT_CHUNK_SIZE,
        resample_to=SR44100(),
        store_resampled=False):
    """
    Create a basic processing pipeline that can resample all incoming audio
    to a normalized sampling rate for downstream processing, and store a
    convenient, compressed version for playback

    :param chunksize_bytes: The number of bytes from the raw stream to process
    at once
    :param resample_to: The new, normalized sampling rate
    :return: A simple processing pipeline
    """

    class Resampled(BaseModel):
        meta = JSONFeature(
            MetaData,
            store=True,
            encoder=AudioMetaDataEncoder)

        raw = ByteStreamFeature(
            ByteStream,
            chunksize=chunksize_bytes,
            needs=meta,
            store=False)

        ogg = OggVorbisFeature(
            OggVorbis,
            needs=raw,
            store=True)

        pcm = AudioSamplesFeature(
            AudioStream,
            needs=raw,
            store=False)

        resampled = AudioSamplesFeature(
            Resampler,
            needs=pcm,
            samplerate=resample_to,
            store=store_resampled)

    return Resampled


def windowed(
        wscheme,
        chunksize_bytes=DEFAULT_CHUNK_SIZE,
        resample_to=SR44100(),
        store_resampled=True,
        store_windowed=False,
        wfunc=None):

    rs = resampled(
        chunksize_bytes=chunksize_bytes,
        resample_to=resample_to,
        store_resampled=store_resampled)

    class Sound(rs):
        windowed = ArrayWithUnitsFeature(
            SlidingWindow,
            wscheme=wscheme,
            wfunc=wfunc,
            needs=rs.resampled,
            store=store_windowed)

    return Sound


def frequency_adaptive(
        long_window_sample_rate,
        scale,
        store_freq_adaptive=False,
        check_scale_overlap_ratio=False,
        chunksize_bytes=DEFAULT_CHUNK_SIZE,
        resample_to=SR44100(),
        store_resampled=False):
    BaseModel = resampled(chunksize_bytes, resample_to, store_resampled)

    class FrequencyAdaptive(BaseModel):
        long_windowed = ArrayWithUnitsFeature(
            SlidingWindow,
            wscheme=long_window_sample_rate,
            wfunc=OggVorbisWindowingFunc(),
            needs=BaseModel.resampled,
            store=False)

        long_fft = ArrayWithUnitsFeature(
            FFT,
            needs=long_windowed,
            store=False)

        freq_adaptive = FrequencyAdaptiveFeature(
            FrequencyAdaptiveTransform,
            transform=np.fft.irfft,
            scale=scale,
            check_scale_overlap_ratio=check_scale_overlap_ratio,
            window_func=np.hanning,
            needs=long_fft,
            store=store_freq_adaptive)

    return FrequencyAdaptive


def stft(
        chunksize_bytes=DEFAULT_CHUNK_SIZE,
        resample_to=SR44100(),
        wscheme=HalfLapped(),
        store_fft=False,
        fft_padding_samples=None,
        store_windowed=False,
        store_resampled=False):
    class ShortTimeFourierTransform(BaseModel):
        meta = JSONFeature(
            MetaData,
            store=True,
            encoder=AudioMetaDataEncoder)

        raw = ByteStreamFeature(
            ByteStream,
            chunksize=chunksize_bytes,
            needs=meta,
            store=False)

        ogg = OggVorbisFeature(
            OggVorbis,
            needs=raw,
            store=True)

        pcm = AudioSamplesFeature(
            AudioStream,
            needs=raw,
            store=False)

        resampled = AudioSamplesFeature(
            Resampler,
            needs=pcm,
            samplerate=resample_to,
            store=store_resampled)

        windowed = ArrayWithUnitsFeature(
            SlidingWindow,
            needs=resampled,
            wscheme=wscheme,
            wfunc=OggVorbisWindowingFunc(),
            store=store_windowed)

        fft = ArrayWithUnitsFeature(
            FFT,
            padding_samples=fft_padding_samples,
            needs=windowed,
            store=store_fft)

    return ShortTimeFourierTransform


def audio_graph(
        chunksize_bytes=DEFAULT_CHUNK_SIZE,
        resample_to=SR44100(),
        store_fft=False):
    """
    Produce a base class suitable as a starting point for many audio processing
    pipelines.  This class resamples all audio to a common sampling rate, and
    produces a bark band spectrogram from overlapping short-time fourier
    transform frames.  It also compresses the audio into ogg vorbis format for
    compact storage.
    """

    band = FrequencyBand(20, resample_to.nyquist)

    class AudioGraph(BaseModel):
        meta = JSONFeature(
            MetaData,
            store=True,
            encoder=AudioMetaDataEncoder)

        raw = ByteStreamFeature(
            ByteStream,
            chunksize=chunksize_bytes,
            needs=meta,
            store=False)

        ogg = OggVorbisFeature(
            OggVorbis,
            needs=raw,
            store=True)

        pcm = AudioSamplesFeature(
            AudioStream,
            needs=raw,
            store=False)

        resampled = AudioSamplesFeature(
            Resampler,
            needs=pcm,
            samplerate=resample_to,
            store=False)

        windowed = ArrayWithUnitsFeature(
            SlidingWindow,
            needs=resampled,
            wscheme=HalfLapped(),
            wfunc=OggVorbisWindowingFunc(),
            store=False)

        dct = ArrayWithUnitsFeature(
            DCT,
            needs=windowed,
            store=True)

        fft = ArrayWithUnitsFeature(
            FFT,
            needs=windowed,
            store=store_fft)

        bark = ArrayWithUnitsFeature(
            BarkBands,
            needs=fft,
            frequency_band=band,
            store=True)

        centroid = ArrayWithUnitsFeature(
            SpectralCentroid,
            needs=bark,
            store=True)

        chroma = ArrayWithUnitsFeature(
            Chroma,
            needs=fft,
            frequency_band=band,
            store=True)

        bfcc = ArrayWithUnitsFeature(
            BFCC,
            needs=fft,
            store=True)

    return AudioGraph


def with_onsets(fft_feature):
    """
    Produce a mixin class that extracts onsets
    :param fft_feature: The short-time fourier transform feature
    :return: A mixin class that extracts onsets
    """

    class Onsets(BaseModel):
        onset_prep = ArrayWithUnitsFeature(
            SlidingWindow,
            needs=fft_feature,
            wscheme=HalfLapped() * Stride(frequency=1, duration=3),
            store=False)

        complex_domain = ArrayWithUnitsFeature(
            ComplexDomain,
            needs=onset_prep,
            store=False)

        sliding_detection = ArrayWithUnitsFeature(
            SlidingWindow,
            needs=complex_domain,
            wscheme=HalfLapped() * Stride(frequency=1, duration=11),
            padwith=5,
            store=False)

        slices = TimeSliceFeature(
            MovingAveragePeakPicker,
            needs=sliding_detection,
            aggregate=np.median,
            store=True)

    return Onsets



import featureflow
import numpy as np
import unittest2

from preprocess import \
    UnitNorm, MeanStdNormalization, Binarize, PreprocessingPipeline, Log
from zounds.spectral import \
    GeometricScale, FrequencyAdaptive
from zounds.timeseries import Seconds, TimeDimension
from zounds.util import simple_in_memory_settings


class TestPipeline(unittest2.TestCase):
    def test_cannot_invert_pipeline_if_any_steps_are_missing(self):
        class Settings(featureflow.PersistenceSettings):
            id_provider = featureflow.UuidProvider()
            key_builder = featureflow.StringDelimitedKeyBuilder()
            database = featureflow.InMemoryDatabase(key_builder=key_builder)

        class Model(featureflow.BaseModel, Settings):
            unitnorm = featureflow.PickleFeature(
                UnitNorm,
                store=False)

            binary = featureflow.PickleFeature(
                Binarize,
                needs=unitnorm,
                store=False)

            pipeline = featureflow.PickleFeature(
                PreprocessingPipeline,
                needs=(unitnorm, binary),
                store=True)

        data = np.random.random_sample((1000, 4))
        _id = Model.process(unitnorm=data)
        example = np.random.random_sample((10, 4))
        model = Model(_id)
        transformed = model.pipeline.transform(example)
        self.assertRaises(
            NotImplementedError, lambda: transformed.inverse_transform())

    def test_can_invert_pipeline(self):
        class Settings(featureflow.PersistenceSettings):
            id_provider = featureflow.UuidProvider()
            key_builder = featureflow.StringDelimitedKeyBuilder()
            database = featureflow.InMemoryDatabase(key_builder=key_builder)

        class Model(featureflow.BaseModel, Settings):
            unitnorm = featureflow.PickleFeature(
                UnitNorm,
                store=False)

            meanstd = featureflow.PickleFeature(
                MeanStdNormalization,
                needs=unitnorm,
                store=False)

            pipeline = featureflow.PickleFeature(
                PreprocessingPipeline,
                needs=(unitnorm, meanstd),
                store=True)

        data = np.random.random_sample((1000, 4))
        _id = Model.process(unitnorm=data)
        example = np.random.random_sample((10, 4))
        model = Model(_id)
        transformed = model.pipeline.transform(example)
        reconstructed = transformed.inverse_transform()
        diff = np.abs(example - reconstructed)
        self.assertTrue(np.all(diff < .00001))

    def test_can_invert_pipeline_with_log(self):
        class Settings(featureflow.PersistenceSettings):
            id_provider = featureflow.UuidProvider()
            key_builder = featureflow.StringDelimitedKeyBuilder()
            database = featureflow.InMemoryDatabase(key_builder=key_builder)

        class Model(featureflow.BaseModel, Settings):
            log = featureflow.PickleFeature(
                Log,
                store=False)

            meanstd = featureflow.PickleFeature(
                MeanStdNormalization,
                needs=log,
                store=False)

            pipeline = featureflow.PickleFeature(
                PreprocessingPipeline,
                needs=(log, meanstd),
                store=True)

        data = np.random.random_sample((1000, 4))
        _id = Model.process(log=data)
        example = np.random.random_sample((10, 4))
        model = Model(_id)
        transformed = model.pipeline.transform(example)
        reconstructed = transformed.inverse_transform()
        diff = np.abs(example - reconstructed)
        self.assertTrue(np.all(diff < .00001))

    def test_can_invert_pipeline_that_takes_frequency_adaptive_transform(self):
        td = TimeDimension(frequency=Seconds(1))
        scale = GeometricScale(20, 5000, 0.05, 10)
        arrs = [np.zeros((10, x)) for x in xrange(1, 11)]
        fa = FrequencyAdaptive(arrs, td, scale)

        @simple_in_memory_settings
        class Model(featureflow.BaseModel):
            log = featureflow.PickleFeature(
                Log,
                store=False)

            meanstd = featureflow.PickleFeature(
                MeanStdNormalization,
                needs=log,
                store=False)

            pipeline = featureflow.PickleFeature(
                PreprocessingPipeline,
                needs=(log, meanstd),
                store=True)

        _id = Model.process(log=fa)
        model = Model(_id)
        result = model.pipeline.transform(fa)
        recon = result.inverse_transform()
        self.assertIsInstance(recon, FrequencyAdaptive)
        self.assertEqual(fa.shape, recon.shape)


from setuptools import setup
import re
import os

try:
    import pypandoc

    long_description = pypandoc.convert('README.md', 'rst')
except(IOError, ImportError):
    long_description = open('README.md').read()

with open('zounds/__init__.py', 'r') as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
        fd.read(),
        re.MULTILINE).group(1)

download_url = 'https://github.com/jvinyard/zounds/tarball/{version}' \
    .format(**locals())

on_rtd = os.environ.get('READTHEDOCS') == 'True'

extension_modules = []

if not on_rtd:
    try:
        import numpy as np
        from distutils.extension import Extension

        countbits = Extension(
            name='countbits',
            sources=['zounds/nputil/countbits.pyx'],
            include_dirs=[np.get_include()],
            extra_compile_args=[
                '-shared',
                '-pthread',
                '-fPIC',
                '-fwrapv',
                '-O2',
                '-Wall',
                '-fno-strict-aliasing'
            ])
        extension_modules = [countbits]
    except ImportError:
        extension_modules = []

setup(
    name='zounds',
    version=version,
    url='https://github.com/JohnVinyard/zounds',
    author='John Vinyard',
    author_email='john.vinyard@gmail.com',
    long_description=long_description,
    download_url=download_url,
    ext_modules=extension_modules,
    packages=[
        'zounds',
        'zounds.basic',
        'zounds.datasets',
        'zounds.index',
        'zounds.learn',
        'zounds.nputil',
        'zounds.segment',
        'zounds.soundfile',
        'zounds.spectral',
        'zounds.loudness',
        'zounds.synthesize',
        'zounds.timeseries',
        'zounds.ui',
        'zounds.persistence',
        'zounds.core',
        'zounds.util'
    ],
    install_requires=[
        'featureflow',
        'nose',
        'unittest2',
        'certifi==2017.7.27.1',
        'requests',
        'tornado',
        'pysoundfile',
        'matplotlib',
        'argparse',
        'intervaltree',
        'ujson'
    ],
    package_data={
        'nputil': ['*.pyx', '*.pyxbld'],
        'ui': ['*.html', '*.js']
    },
    scripts=['bin/zounds-quickstart'],
    include_package_data=True
)



import featureflow as ff
from util import from_var
from random import choice
from zounds.spectral import stft, rainbowgram
from zounds.learn import \
    try_network, infinite_streaming_learning_pipeline, \
    object_store_pipeline_settings
from zounds.timeseries import \
    SR11025, SampleRate, Seconds, AudioSamples, audio_sample_rate
from wgan import WassersteinGanTrainer
from pytorch_model import PyTorchGan
from preprocess import InstanceScaling
from zounds.ui import GanTrainingMonitorApp
from zounds.util import simple_lmdb_settings
from zounds.basic import windowed
from zounds.spectral import HanningWindowingFunc
from zounds.datasets import ingest
import numpy as np


class GanExperiment(object):
    def __init__(
            self,
            experiment_name,
            dataset,
            gan_pair,
            object_storage_username,
            object_storage_api_key,
            epochs=500,
            n_critic_iterations=10,
            batch_size=32,
            n_samples=int(5e5),
            latent_dim=100,
            real_sample_transformer=lambda x: x,
            debug_gradients=False,
            sample_size=8192,
            sample_hop=1024,
            samplerate=SR11025(),
            app_port=8888,
            object_storage_region='DFW',
            app_secret=None):

        super(GanExperiment, self).__init__()
        self.real_sample_transformer = real_sample_transformer
        self.debug_gradients = debug_gradients
        self.n_samples = n_samples
        self.batch_size = batch_size
        self.n_critic_iterations = n_critic_iterations
        self.epochs = epochs
        self.gan_pair = gan_pair
        self.app_port = app_port
        self.dataset = dataset

        self.samplerate = samplerate
        self.sample_hop = sample_hop
        self.sample_size = sample_size
        self.latent_dim = latent_dim
        self.experiment_name = experiment_name
        self.app_secret = app_secret

        base_model = windowed(
            resample_to=self.samplerate,
            store_resampled=True,
            wscheme=self.samplerate * (sample_hop, sample_size))

        @simple_lmdb_settings(
            experiment_name, map_size=1e11, user_supplied_id=True)
        class Sound(base_model):
            pass

        self.sound_cls = Sound

        @object_store_pipeline_settings(
            'Gan-{experiment_name}'.format(**locals()),
            object_storage_region,
            object_storage_username,
            object_storage_api_key)
        @infinite_streaming_learning_pipeline
        class Gan(ff.BaseModel):
            scaled = ff.PickleFeature(
                InstanceScaling)

            wgan = ff.PickleFeature(
                PyTorchGan,
                trainer=ff.Var('trainer'),
                needs=scaled)

        self.gan_pipeline = Gan()
        self.fake_samples = None
        self.app = None

    def batch_complete(self, *args, **kwargs):
        samples = kwargs['samples']
        self.fake_samples = from_var(samples).squeeze()

    def fake_audio(self):
        sample = choice(self.fake_samples)
        return AudioSamples(sample, self.samplerate) \
            .pad_with_silence(Seconds(1))

    def _stft(self, samples):
        samples = samples / np.abs(samples.max())
        wscheme = SampleRate(
            frequency=samples.samplerate.frequency * 128,
            duration=samples.samplerate.frequency * 256)
        coeffs = stft(samples, wscheme, HanningWindowingFunc())
        return rainbowgram(coeffs)

    def fake_stft(self):
        samples = self.fake_audio()
        return self._stft(samples)

    def real_stft(self):
        snd = self.sound_cls.random()
        windowed = choice(snd.windowed)
        windowed = AudioSamples(
            windowed,
            audio_sample_rate(windowed.dimensions[0].samples_per_second))
        return self._stft(windowed)

    def test(self):
        z = np.random.normal(
            0, 1, (self.batch_size, self.latent_dim)).astype(np.float32)
        samples = try_network(self.gan_pair.generator, z)
        samples = from_var(samples)
        print samples.shape
        wasserstein_estimate = try_network(self.gan_pair.discriminator, samples)
        print wasserstein_estimate.shape

    def run(self):
        ingest(self.dataset, self.sound_cls, multi_threaded=True)

        experiment = self
        fake_audio = self.fake_audio
        fake_stft = self.fake_stft
        real_stft = self.real_stft
        Sound = self.sound_cls

        try:
            network = self.gan_pipeline.load_network()
            print 'initialized weights'
        except RuntimeError as e:
            print 'Error', e
            network = self.gan_pair
            for p in network.parameters():
                p.data.normal_(0, 0.02)

        trainer = WassersteinGanTrainer(
            network,
            latent_dimension=(self.latent_dim,),
            n_critic_iterations=self.n_critic_iterations,
            epochs=self.epochs,
            batch_size=self.batch_size,
            debug_gradient=self.debug_gradients)
        trainer.register_batch_complete_callback(self.batch_complete)

        self.app = GanTrainingMonitorApp(
            trainer=trainer,
            model=Sound,
            visualization_feature=Sound.windowed,
            audio_feature=Sound.ogg,
            globals=globals(),
            locals=locals(),
            secret=self.app_secret)

        with self.app.start_in_thread(self.app_port):
            self.gan_pipeline.process(
                dataset=(Sound, Sound.windowed),
                trainer=trainer,
                nsamples=self.n_samples,
                dtype=np.float32)

        self.app.start(self.app_port)



from trainer import Trainer
import numpy as np


class WassersteinGanTrainer(Trainer):
    """
    Args:
        network (nn.Module): the network to train
        latent_dimension (tuple): A tuple that defines the shape of the latent
            dimension (noise) that is the generator's input
        n_critic_iterations (int): The number of minibatches the critic sees
            for every minibatch the generator sees
        epochs: The total number of passes over the training set
        batch_size: The size of a minibatch
        preprocess_minibatch (function): function that takes the current
            epoch, and a minibatch, and mutates the minibatch
        kwargs_factory (callable): function that takes the current epoch and
            outputs args to pass to the generator and discriminator
    """

    def __init__(
            self,
            network,
            latent_dimension,
            n_critic_iterations,
            epochs,
            batch_size,
            preprocess_minibatch=None,
            kwargs_factory=None,
            debug_gradient=False,
            checkpoint_epochs=1):

        super(WassersteinGanTrainer, self).__init__(epochs, batch_size)
        self.checkpoint_epochs = checkpoint_epochs
        self.debug_gradient = debug_gradient
        self.arg_maker = kwargs_factory
        self.preprocess = preprocess_minibatch
        self.n_critic_iterations = n_critic_iterations
        self.latent_dimension = latent_dimension
        self.network = network
        self.critic = network.discriminator
        self.generator = network.generator
        self.samples = None
        self.register_batch_complete_callback(self._log)
        self.generator_optim = None
        self.critic_optim = None

    def _log(self, *args, **kwargs):
        if kwargs['batch'] % 10:
            return
        msg = 'Epoch {epoch}, batch {batch}, generator {generator_score}, ' \
              'real {real_score}, critic {critic_loss}'
        print msg.format(**kwargs)

    def _minibatch(self, data):
        indices = np.random.randint(0, len(data), self.batch_size)
        return data[indices, ...]

    def _gradient_penalty(self, real_samples, fake_samples, kwargs):
        """
        Compute the norm of the gradients for each sample in a batch, and
        penalize anything on either side of unit norm
        """
        import torch
        from torch.autograd import Variable, grad

        real_samples = real_samples.view(fake_samples.shape)

        subset_size = real_samples.shape[0]

        real_samples = real_samples[:subset_size]
        fake_samples = fake_samples[:subset_size]

        alpha = torch.rand(subset_size).cuda()
        alpha = alpha.view((-1,) + ((1,) * (real_samples.dim() - 1)))

        interpolates = alpha * real_samples + ((1 - alpha) * fake_samples)
        interpolates = Variable(interpolates.cuda(), requires_grad=True)

        d_output = self.critic(interpolates, **kwargs)

        gradients = grad(
            outputs=d_output,
            inputs=interpolates,
            grad_outputs=torch.ones(d_output.size()).cuda(),
            create_graph=True,
            retain_graph=True,
            only_inputs=True)[0]
        return ((gradients.norm(2, dim=1) - 1) ** 2).mean() * 10

    def freeze_generator(self):
        for p in self.generator.parameters():
            p.requires_grad = False

    def unfreeze_generator(self):
        for p in self.generator.parameters():
            p.requires_grad = True

    def freeze_discriminator(self):
        for p in self.critic.parameters():
            p.requires_grad = False

    def unfreeze_discriminator(self):
        for p in self.critic.parameters():
            p.requires_grad = True

    def _debug_network_gradient(self, network):
        if not self.debug_gradient:
            return

        for n, p in network.named_parameters():
            g = p.grad
            if g is not None:
                print(n, g.min().data[0], g.max().data[0], g.mean().data[0])

    def zero_generator_gradients(self):
        self._debug_network_gradient(self.generator)
        self.generator.zero_grad()

    def zero_discriminator_gradients(self):
        self._debug_network_gradient(self.critic)
        self.critic.zero_grad()

    def _init_optimizers(self):
        if self.generator_optim is None or self.critic_optim is None:
            from torch.optim import Adam
            trainable_generator_params = (
                p for p in self.generator.parameters() if p.requires_grad)
            trainable_critic_params = (
                p for p in self.critic.parameters() if p.requires_grad)

            self.generator_optim = Adam(
                trainable_generator_params, lr=0.0001, betas=(0, 0.9))
            self.critic_optim = Adam(
                trainable_critic_params, lr=0.0001, betas=(0, 0.9))

    def train(self, data):

        import torch
        from torch.autograd import Variable

        self.network.train()
        self.unfreeze_discriminator()
        self.unfreeze_generator()

        data = data.astype(np.float32)

        noise_shape = (self.batch_size,) + self.latent_dimension
        noise = torch.FloatTensor(*noise_shape)
        fixed_noise = torch.FloatTensor(*noise_shape).normal_(0, 1)

        self.generator.cuda()
        self.critic.cuda()
        noise, fixed_noise = noise.cuda(), fixed_noise.cuda()

        self._init_optimizers()

        start = self._current_epoch
        stop = self._current_epoch + self.checkpoint_epochs

        for epoch in xrange(start, stop):
            if epoch >= self.epochs:
                break

            if self.arg_maker:
                kwargs = self.arg_maker(epoch)
            else:
                kwargs = dict()

            for i in xrange(0, len(data), self.batch_size):
                self.zero_generator_gradients()
                self.zero_discriminator_gradients()

                self.freeze_generator()
                self.unfreeze_discriminator()

                for c in xrange(self.n_critic_iterations):

                    self.zero_discriminator_gradients()

                    minibatch = self._minibatch(data)
                    inp = torch.from_numpy(minibatch)
                    inp = inp.cuda()
                    input_v = Variable(inp)

                    if self.preprocess:
                        input_v = self.preprocess(epoch, input_v)

                    d_real = self.critic.forward(input_v, **kwargs)

                    # train discriminator on fake data
                    noise.normal_(0, 1)
                    noise_v = Variable(noise, volatile=True)
                    fake = Variable(
                        self.generator.forward(noise_v, **kwargs).data)

                    if self.preprocess:
                        fake = self.preprocess(epoch, fake)

                    d_fake = self.critic.forward(fake, **kwargs)

                    real_mean = torch.mean(d_real)
                    fake_mean = torch.mean(d_fake)
                    gp = self._gradient_penalty(input_v.data, fake.data, kwargs)
                    d_loss = (fake_mean - real_mean) + gp
                    d_loss.backward()
                    self.critic_optim.step()

                self.zero_discriminator_gradients()
                self.zero_generator_gradients()

                self.unfreeze_generator()
                self.freeze_discriminator()

                # train generator
                noise.normal_(0, 1)
                noise_v = Variable(noise)
                fake = self.generator.forward(noise_v, **kwargs)

                if self.preprocess:
                    fake = self.preprocess(epoch, fake)

                self.samples = fake

                d_fake = self.critic.forward(fake, **kwargs)
                g_loss = -torch.mean(d_fake)
                g_loss.backward()
                self.generator_optim.step()

                gl = g_loss.data[0]
                dl = d_loss.data[0]
                rl = real_mean.data[0]

                self.on_batch_complete(
                    epoch=epoch,
                    batch=i,
                    generator_score=gl,
                    real_score=rl,
                    critic_loss=dl,
                    samples=self.samples,
                    network=self.network)

            self._current_epoch += 1

        return self.network



from zounds.timeseries import AudioSamples
from resample import Resampler
import numpy as np


def resample(samples, new_sample_rate):
    if new_sample_rate == samples.samplerate:
        return samples
    rs = Resampler(new_sample_rate)
    new_samples = np.concatenate(list(rs._process(samples)))
    return AudioSamples(new_samples, new_sample_rate)



from timeseries import TimeDimension, TimeSlice
from zounds.core import ArrayWithUnits


class ConstantRateTimeSeries(ArrayWithUnits):
    def __new__(cls, array):
        try:
            dim = array.dimensions[0]
        except AttributeError:
            raise ValueError('array must be of type ArrayWithUnits')

        if not isinstance(dim, TimeDimension):
            raise ValueError('array first dimension must be a TimeDimension')

        return ArrayWithUnits.__new__(cls, array, array.dimensions)

    @property
    def time_dimension(self):
        return self.dimensions[0]

    def iter_slices(self):
        td = self.dimensions[0]
        for i, data in enumerate(self):
            yield TimeSlice(duration=td.duration, start=td.frequency * i), data



class Dimension(object):
    """
    Common base class representing one dimension of a numpy array.  Sub-classes
    can define behavior making custom slices (e.g., time spans or
    frequency bands) possible.

    Implementors are primarily responsible for determining how custom slices
    are transformed into integer indexes and slices that numpy can use directly.

    See Also:
        :class:`IdentityDimension`
        :class:`~zounds.timeseries.TimeDimension`
        :class:`~zounds.spectral.FrequencyDimension`
    """

    def __init__(self):
        super(Dimension, self).__init__()

    def modified_dimension(self, size, windowsize, stepsize=None):
        raise NotImplementedError()

    def metaslice(self, index, size):
        """
        Produce a new instance of this dimension, given a custom slice
        """
        return self

    def integer_based_slice(self, index):
        """
        Subclasses define behavior that transforms a custom, user-defined slice
        into integer indices that numpy can understand

        Args:
            index (custom slice): A user-defined slice instance
        """
        raise NotImplementedError()

    def validate(self, size):
        """
        Subclasses check to ensure that the dimensions size does not validate
        any assumptions made by this instance
        """
        pass

    def copy(self):
        return Dimension()


class IdentityDimension(Dimension):
    """
    A custom dimension that does not transform indices in any way, simply acting
    as a pass-through.

    Examples:
        >>> from zounds import ArrayWithUnits, IdentityDimension
        >>> import numpy as np
        >>> data = np.zeros(100)
        >>> arr = ArrayWithUnits(data, [IdentityDimension()])
        >>> sliced = arr[4:6]
        >>> sliced.shape
        (2,)
    """

    def __init__(self):
        super(IdentityDimension, self).__init__()

    def modified_dimension(self, size, windowsize, stepsize=None):
        if windowsize == slice(None) or (size / windowsize) == 1:
            yield IdentityDimension()
        else:
            raise ValueError()

    def integer_based_slice(self, index):
        return index

    def __eq__(self, other):
        return self.__class__ == other.__class__

    def copy(self):
        return IdentityDimension()



import unittest2
from ogg_vorbis import OggVorbisWrapper
from zounds.timeseries import TimeSlice, Seconds, SR11025
from zounds.synthesize import SineSynthesizer


class TestOggVorbisWrapper(unittest2.TestCase):
    def test_can_apply_empty_time_slice_to_wrapper(self):
        synth = SineSynthesizer(SR11025())
        samples = synth.synthesize(Seconds(10))
        encoded = samples.encode(fmt='OGG', subtype='VORBIS')
        wrapper = OggVorbisWrapper(encoded)
        samples = wrapper[TimeSlice()]
        expected = Seconds(10) / Seconds(1)
        actual = samples.end / Seconds(1)
        self.assertAlmostEqual(expected, actual, places=6)

    def test_can_apply_open_ended_slice_to_wrapper(self):
        synth = SineSynthesizer(SR11025())
        samples = synth.synthesize(Seconds(10))
        encoded = samples.encode(fmt='OGG', subtype='VORBIS')
        wrapper = OggVorbisWrapper(encoded)
        samples = wrapper[TimeSlice(start=Seconds(1))]
        expected = Seconds(9) / Seconds(1)
        actual = samples.end / Seconds(1)
        self.assertAlmostEqual(expected, actual, places=6)
